@@ -43,7 +43,6 @@ import (
 type typeDetails struct {
 	outputType         bool
 	inputType          bool
-	argsType           bool
 	resourceOutputType bool
 	plainType          bool
 }
@@ -141,7 +140,7 @@ func (mod *modContext) details(t *schema.ObjectType) *typeDetails {
 	return details
 }
 
-func (mod *modContext) modNameAndName(pkg *schema.Package, t schema.Type, input, args bool) (modName string, name string) {
+func (mod *modContext) modNameAndName(pkg *schema.Package, t schema.Type, input bool) (modName string, name string) {
 	var info PackageInfo
 	contract.AssertNoError(pkg.ImportLanguages(map[string]schema.Language{"python": Importer}))
 	if v, ok := pkg.Language["python"].(PackageInfo); ok {
@@ -158,7 +157,7 @@ func (mod *modContext) modNameAndName(pkg *schema.Package, t schema.Type, input,
 			modNameOverrides: info.ModuleNameOverrides,
 			compatibility:    info.Compatibility,
 		}
-		token, name = t.Token, namingCtx.unqualifiedObjectTypeName(t, input, args)
+		token, name = t.Token, namingCtx.unqualifiedObjectTypeName(t, input)
 	case *schema.ResourceType:
 		token, name = t.Token, tokenToName(t.Token)
 	}
@@ -173,11 +172,11 @@ func (mod *modContext) modNameAndName(pkg *schema.Package, t schema.Type, input,
 	return
 }
 
-func (mod *modContext) unqualifiedObjectTypeName(t *schema.ObjectType, input, args bool) string {
+func (mod *modContext) unqualifiedObjectTypeName(t *schema.ObjectType, input bool) string {
 	name := tokenToName(t.Token)
 
 	if mod.compatibility != tfbridge20 && mod.compatibility != kubernetes20 {
-		if args {
+		if t.IsInputShape() {
 			return name + "Args"
 		}
 		return name
@@ -192,7 +191,7 @@ func (mod *modContext) unqualifiedObjectTypeName(t *schema.ObjectType, input, ar
 	return name
 }
 
-func (mod *modContext) objectType(t *schema.ObjectType, input, args bool) string {
+func (mod *modContext) objectType(t *schema.ObjectType, input bool) string {
 	var prefix string
 	if !input {
 		prefix = "outputs."
@@ -200,11 +199,11 @@ func (mod *modContext) objectType(t *schema.ObjectType, input, args bool) string
 
 	// If it's an external type, reference it via fully qualified name.
 	if t.Package != mod.pkg {
-		modName, name := mod.modNameAndName(t.Package, t, input, args)
+		modName, name := mod.modNameAndName(t.Package, t, input)
 		return fmt.Sprintf("'%s.%s%s%s'", pyPack(t.Package.Name), modName, prefix, name)
 	}
 
-	modName, name := mod.tokenToModule(t.Token), mod.unqualifiedObjectTypeName(t, input, args)
+	modName, name := mod.tokenToModule(t.Token), mod.unqualifiedObjectTypeName(t, input)
 	if modName == "" && modName != mod.mod {
 		rootModName := "_root_outputs."
 		if input {
@@ -252,7 +251,7 @@ func (mod *modContext) resourceType(r *schema.ResourceType) string {
 	}
 
 	pkg := r.Resource.Package
-	modName, name := mod.modNameAndName(pkg, r, false, false)
+	modName, name := mod.modNameAndName(pkg, r, false)
 	return fmt.Sprintf("%s.%s%s", pyPack(pkg.Name), modName, name)
 }
 
@@ -291,10 +290,18 @@ func tokenToName(tok string) string {
 }
 
 func tokenToModule(tok string, pkg *schema.Package, moduleNameOverrides map[string]string) string {
+	// See if there's a manually-overridden module name.
 	canonicalModName := pkg.TokenToModule(tok)
-	modName := PyName(strings.ToLower(canonicalModName))
 	if override, ok := moduleNameOverrides[canonicalModName]; ok {
-		modName = override
+		return override
+	}
+	// A module can include fileparts, which we want to preserve.
+	var modName string
+	for i, part := range strings.Split(strings.ToLower(canonicalModName), "/") {
+		if i > 0 {
+			modName += "/"
+		}
+		modName += PyName(part)
 	}
 	return modName
 }
@@ -524,9 +531,9 @@ func (mod *modContext) unqualifiedImportName() string {
 
 	// Extract version suffix from child modules. Nested versions will have their own __init__.py file.
 	// Example: apps/v1beta1 -> v1beta1
-	parts := strings.SplitN(name, "/", 2)
-	if len(parts) == 2 {
-		name = parts[1]
+	parts := strings.Split(name, "/")
+	if len(parts) > 1 {
+		name = parts[len(parts)-1]
 	}
 
 	return PyName(name)
@@ -746,7 +753,7 @@ func (mod *modContext) genConfig(variables []*schema.Property) (string, error) {
 	for _, p := range variables {
 		configFetch := fmt.Sprintf("__config__.get('%s')", p.Name)
 		if p.DefaultValue != nil {
-			v, err := getDefaultValue(p.DefaultValue, p.Type)
+			v, err := getDefaultValue(p.DefaultValue, codegen.UnwrapType(p.Type))
 			if err != nil {
 				return "", err
 			}
@@ -768,7 +775,7 @@ func (mod *modContext) genTypes(dir string, fs fs) error {
 		imports := imports{}
 		for _, t := range mod.types {
 			if input && mod.details(t).inputType {
-				visitObjectTypes(t.Properties, func(t schema.Type, _ bool) {
+				visitObjectTypes(t.Properties, func(t schema.Type) {
 					switch t := t.(type) {
 					case *schema.ObjectType:
 						imports.addTypeIf(mod, t, true /*input*/, func(imp string) bool {
@@ -795,15 +802,8 @@ func (mod *modContext) genTypes(dir string, fs fs) error {
 		// Export only the symbols we want exported.
 		fmt.Fprintf(w, "__all__ = [\n")
 		for _, t := range mod.types {
-			if input && mod.details(t).inputType {
-				if mod.details(t).argsType {
-					fmt.Fprintf(w, "    '%s',\n", mod.unqualifiedObjectTypeName(t, true, true))
-				}
-				if mod.details(t).plainType {
-					fmt.Fprintf(w, "    '%s',\n", mod.unqualifiedObjectTypeName(t, true, false))
-				}
-			} else if !input && mod.details(t).outputType {
-				fmt.Fprintf(w, "    '%s',\n", mod.unqualifiedObjectTypeName(t, false, false))
+			if input && mod.details(t).inputType || !input && mod.details(t).outputType {
+				fmt.Fprintf(w, "    '%s',\n", mod.unqualifiedObjectTypeName(t, input))
 			}
 		}
 		fmt.Fprintf(w, "]\n\n")
@@ -878,8 +878,7 @@ func (mod *modContext) genAwaitableType(w io.Writer, obj *schema.ObjectType) str
 
 	// Write out Python property getters for each property.
 	mod.genProperties(w, obj.Properties, false /*setters*/, func(prop *schema.Property) string {
-		return mod.typeString(prop.Type, false /*input*/, false /*wrapInput*/, false /*args*/, !prop.IsRequired,
-			false /*acceptMapping*/)
+		return mod.typeString(prop.Type, false /*input*/, false /*acceptMapping*/)
 	})
 
 	// Produce an awaitable subclass.
@@ -935,7 +934,7 @@ func (mod *modContext) genResource(res *schema.Resource) (string, error) {
 	const alternateSuffix = "InitArgs"
 	for _, t := range mod.types {
 		if mod.details(t).inputType {
-			if mod.unqualifiedObjectTypeName(t, true, true) == resourceArgsName {
+			if mod.unqualifiedObjectTypeName(t, true) == resourceArgsName {
 				resourceArgsName = name + alternateSuffix
 				break
 			}
@@ -945,7 +944,7 @@ func (mod *modContext) genResource(res *schema.Resource) (string, error) {
 	if strings.HasSuffix(resourceArgsName, alternateSuffix) {
 		for _, t := range mod.types {
 			if mod.details(t).inputType {
-				if mod.unqualifiedObjectTypeName(t, true, true) == resourceArgsName {
+				if mod.unqualifiedObjectTypeName(t, true) == resourceArgsName {
 					return "", errors.Errorf(
 						"resource args class named %s in %s conflicts with input type", resourceArgsName, mod.mod)
 				}
@@ -958,7 +957,7 @@ func (mod *modContext) genResource(res *schema.Resource) (string, error) {
 
 	// Produce an args class.
 	argsComment := fmt.Sprintf("The set of arguments for constructing a %s resource.", name)
-	err := mod.genType(w, resourceArgsName, argsComment, res.InputProperties, false, true, true, false)
+	err := mod.genType(w, resourceArgsName, argsComment, res.InputProperties, true, false)
 	if err != nil {
 		return "", err
 	}
@@ -971,7 +970,7 @@ func (mod *modContext) genResource(res *schema.Resource) (string, error) {
 		len(res.StateInputs.Properties) > 0
 	if hasStateInputs {
 		stateComment := fmt.Sprintf("Input properties used for looking up and filtering %s resources.", name)
-		err = mod.genType(w, fmt.Sprintf("_%sState", name), stateComment, res.StateInputs.Properties, false, true, true, false)
+		err = mod.genType(w, fmt.Sprintf("_%sState", name), stateComment, res.StateInputs.Properties, true, false)
 		if err != nil {
 			return "", err
 		}
@@ -1002,7 +1001,7 @@ func (mod *modContext) genResource(res *schema.Resource) (string, error) {
 	// Determine if all inputs are optional.
 	allOptionalInputs := true
 	for _, prop := range res.InputProperties {
-		allOptionalInputs = allOptionalInputs && !prop.IsRequired
+		allOptionalInputs = allOptionalInputs && !prop.IsRequired()
 	}
 
 	// Emit __init__ overloads and implementation...
@@ -1015,8 +1014,7 @@ func (mod *modContext) genResource(res *schema.Resource) (string, error) {
 
 		// If there's an argument type, emit it.
 		for _, prop := range res.InputProperties {
-			wrapInput := !prop.IsPlain
-			ty := mod.typeString(prop.Type, true, wrapInput, wrapInput, true /*optional*/, true /*acceptMapping*/)
+			ty := mod.typeString(codegen.OptionalType(prop), true, true /*acceptMapping*/)
 			fmt.Fprintf(w, ",\n                 %s: %s = None", InitParamName(prop.Name), ty)
 		}
 
@@ -1088,7 +1086,7 @@ func (mod *modContext) genResource(res *schema.Resource) (string, error) {
 
 		// Fill in computed defaults for arguments.
 		if prop.DefaultValue != nil {
-			dv, err := getDefaultValue(prop.DefaultValue, prop.Type)
+			dv, err := getDefaultValue(prop.DefaultValue, codegen.UnwrapType(prop.Type))
 			if err != nil {
 				return "", err
 			}
@@ -1097,7 +1095,7 @@ func (mod *modContext) genResource(res *schema.Resource) (string, error) {
 		}
 
 		// Check that required arguments are present.
-		if prop.IsRequired {
+		if prop.IsRequired() {
 			fmt.Fprintf(w, "            if %s is None and not opts.urn:\n", pname)
 			fmt.Fprintf(w, "                raise TypeError(\"Missing required property '%s'\")\n", pname)
 		}
@@ -1193,8 +1191,8 @@ func (mod *modContext) genResource(res *schema.Resource) (string, error) {
 
 		if hasStateInputs {
 			for _, prop := range res.StateInputs.Properties {
-				pname := PyName(prop.Name)
-				ty := mod.typeString(prop.Type, true, true, true, true /*optional*/, true /*acceptMapping*/)
+				pname := InitParamName(prop.Name)
+				ty := mod.typeString(codegen.OptionalType(prop), true, true /*acceptMapping*/)
 				fmt.Fprintf(w, ",\n            %s: %s = None", pname, ty)
 			}
 		}
@@ -1215,7 +1213,7 @@ func (mod *modContext) genResource(res *schema.Resource) (string, error) {
 		if res.StateInputs != nil {
 			for _, prop := range res.StateInputs.Properties {
 				stateInputs.Add(prop.Name)
-				fmt.Fprintf(w, "        __props__.__dict__[%[1]q] = %[1]s\n", PyName(prop.Name))
+				fmt.Fprintf(w, "        __props__.__dict__[%q] = %s\n", PyName(prop.Name), InitParamName(prop.Name))
 			}
 		}
 		for _, prop := range res.Properties {
@@ -1229,7 +1227,7 @@ func (mod *modContext) genResource(res *schema.Resource) (string, error) {
 
 	// Write out Python property getters for each of the resource's properties.
 	mod.genProperties(w, res.Properties, false /*setters*/, func(prop *schema.Property) string {
-		ty := mod.typeString(prop.Type, false /*input*/, false /*wrapInput*/, false /*args*/, !prop.IsRequired, false /*acceptMapping*/)
+		ty := mod.typeString(prop.Type, false /*input*/, false /*acceptMapping*/)
 		return fmt.Sprintf("pulumi.Output[%s]", ty)
 	})
 
@@ -1312,7 +1310,10 @@ func (mod *modContext) genFunction(fun *schema.Function) (string, error) {
 
 	mod.genHeader(w, true /*needsSDK*/, imports)
 
-	baseName, awaitableName := awaitableTypeNames(fun.Outputs.Token)
+	var baseName, awaitableName string
+	if fun.Outputs != nil {
+		baseName, awaitableName = awaitableTypeNames(fun.Outputs.Token)
+	}
 	name := PyName(tokenToName(fun.Token))
 
 	// Export only the symbols we want exported.
@@ -1355,7 +1356,7 @@ func (mod *modContext) genFunction(fun *schema.Function) (string, error) {
 			ind = indent
 		}
 		pname := PyName(arg.Name)
-		ty := mod.typeString(arg.Type, true, false /*wrapInput*/, false /*args*/, true /*optional*/, true /*acceptMapping*/)
+		ty := mod.typeString(codegen.OptionalType(arg), true, true /*acceptMapping*/)
 		fmt.Fprintf(w, "%s%s: %s = None,\n", ind, pname, ty)
 	}
 	fmt.Fprintf(w, "%sopts: Optional[pulumi.InvokeOptions] = None", indent)
@@ -1375,7 +1376,7 @@ func (mod *modContext) genFunction(fun *schema.Function) (string, error) {
 	if len(args) > 0 {
 		fmt.Fprintln(docs, "")
 		for _, arg := range args {
-			mod.genPropDocstring(docs, PyName(arg.Name), arg, false /*wrapInputs*/, true /*acceptMapping*/)
+			mod.genPropDocstring(docs, PyName(arg.Name), arg, true /*acceptMapping*/)
 		}
 	}
 	printComment(w, docs.String(), "    ")
@@ -1454,7 +1455,7 @@ func (mod *modContext) genEnums(w io.Writer, enums []*schema.EnumType) error {
 func (mod *modContext) genEnum(w io.Writer, enum *schema.EnumType) error {
 	indent := "    "
 	enumName := tokenToName(enum.Token)
-	underlyingType := mod.typeString(enum.ElementType, false, false, false, false, false)
+	underlyingType := mod.typeString(enum.ElementType, false, false)
 
 	switch enum.ElementType {
 	case schema.StringType, schema.IntType, schema.NumberType:
@@ -1478,6 +1479,9 @@ func (mod *modContext) genEnum(w io.Writer, enum *schema.EnumType) error {
 			} else {
 				fmt.Fprintf(w, "%v\n", e.Value)
 			}
+			if e.Comment != "" {
+				fmt.Fprintf(w, "%s\"\"\"%s\"\"\"\n", indent, e.Comment)
+			}
 		}
 	default:
 		return errors.Errorf("enums of type %s are not yet implemented for this language", enum.ElementType.String())
@@ -1486,18 +1490,18 @@ func (mod *modContext) genEnum(w io.Writer, enum *schema.EnumType) error {
 	return nil
 }
 
-func visitObjectTypes(properties []*schema.Property, visitor func(objectOrResource schema.Type, plain bool)) {
-	codegen.VisitTypeClosure(properties, func(t codegen.Type) {
-		switch st := t.Type.(type) {
+func visitObjectTypes(properties []*schema.Property, visitor func(objectOrResource schema.Type)) {
+	codegen.VisitTypeClosure(properties, func(t schema.Type) {
+		switch st := t.(type) {
 		case *schema.EnumType, *schema.ObjectType, *schema.ResourceType:
-			visitor(st, t.Plain)
+			visitor(st)
 		}
 	})
 }
 
 func (mod *modContext) collectImports(properties []*schema.Property, imports imports, input bool) {
-	codegen.VisitTypeClosure(properties, func(t codegen.Type) {
-		switch t := t.Type.(type) {
+	codegen.VisitTypeClosure(properties, func(t schema.Type) {
+		switch t := t.(type) {
 		case *schema.ObjectType:
 			imports.addType(mod, t, input)
 		case *schema.EnumType:
@@ -1582,7 +1586,7 @@ func genPackageMetadata(
 	fmt.Fprintf(w, "        with open('README.md', encoding='utf-8') as f:\n")
 	fmt.Fprintf(w, "            return f.read()\n")
 	fmt.Fprintf(w, "    except FileNotFoundError:\n")
-	fmt.Fprintf(w, "            return \"%s Pulumi Package - Development Version\"\n", pkg.Name)
+	fmt.Fprintf(w, "        return \"%s Pulumi Package - Development Version\"\n", pkg.Name)
 	fmt.Fprintf(w, "\n\n")
 
 	// Finally, the actual setup part.
@@ -1797,7 +1801,7 @@ func (mod *modContext) genInitDocstring(w io.Writer, res *schema.Resource, resou
 	fmt.Fprintln(b, ":param pulumi.ResourceOptions opts: Options for the resource.")
 	if !argOverload {
 		for _, prop := range res.InputProperties {
-			mod.genPropDocstring(b, InitParamName(prop.Name), prop, true /*wrapInput*/, true /*acceptMapping*/)
+			mod.genPropDocstring(b, InitParamName(prop.Name), prop, true /*acceptMapping*/)
 		}
 	}
 
@@ -1818,7 +1822,7 @@ func (mod *modContext) genGetDocstring(w io.Writer, res *schema.Resource) {
 	fmt.Fprintln(b, ":param pulumi.ResourceOptions opts: Options for the resource.")
 	if res.StateInputs != nil {
 		for _, prop := range res.StateInputs.Properties {
-			mod.genPropDocstring(b, PyName(prop.Name), prop, true /*wrapInput*/, true /*acceptMapping*/)
+			mod.genPropDocstring(b, InitParamName(prop.Name), prop, true /*acceptMapping*/)
 		}
 	}
 
@@ -1826,7 +1830,7 @@ func (mod *modContext) genGetDocstring(w io.Writer, res *schema.Resource) {
 	printComment(w, b.String(), "        ")
 }
 
-func (mod *modContext) genTypeDocstring(w io.Writer, comment string, properties []*schema.Property, wrapInput bool) {
+func (mod *modContext) genTypeDocstring(w io.Writer, comment string, properties []*schema.Property) {
 	// b contains the full text of the docstring, without the leading and trailing triple quotes.
 	b := &bytes.Buffer{}
 
@@ -1836,21 +1840,19 @@ func (mod *modContext) genTypeDocstring(w io.Writer, comment string, properties
 	}
 
 	for _, prop := range properties {
-		mod.genPropDocstring(b, PyName(prop.Name), prop, wrapInput, false /*acceptMapping*/)
+		mod.genPropDocstring(b, PyName(prop.Name), prop, false /*acceptMapping*/)
 	}
 
 	// printComment handles the prefix and triple quotes.
 	printComment(w, b.String(), "        ")
 }
 
-func (mod *modContext) genPropDocstring(w io.Writer, name string, prop *schema.Property, wrapInput bool,
-	acceptMapping bool) {
-
+func (mod *modContext) genPropDocstring(w io.Writer, name string, prop *schema.Property, acceptMapping bool) {
 	if prop.Comment == "" {
 		return
 	}
 
-	ty := mod.typeString(prop.Type, true, wrapInput && !prop.IsPlain, wrapInput && !prop.IsPlain, false /*optional*/, acceptMapping)
+	ty := mod.typeString(codegen.RequiredType(prop), true, acceptMapping)
 
 	// If this property has some documentation associated with it, we need to split it so that it is indented
 	// in a way that Sphinx can understand.
@@ -1869,102 +1871,95 @@ func (mod *modContext) genPropDocstring(w io.Writer, name string, prop *schema.P
 	}
 }
 
-func (mod *modContext) typeString(t schema.Type, input, wrapInput, args, optional, acceptMapping bool) string {
-	var typ string
+func (mod *modContext) typeString(t schema.Type, input, acceptMapping bool) string {
 	switch t := t.(type) {
-	case *schema.EnumType:
-		typ = mod.tokenToEnum(t.Token)
-	case *schema.ArrayType:
-		typ = fmt.Sprintf("Sequence[%s]", mod.typeString(t.ElementType, input, wrapInput, args, false, acceptMapping))
-	case *schema.MapType:
-		typ = fmt.Sprintf("Mapping[str, %s]", mod.typeString(t.ElementType, input, wrapInput, args, false, acceptMapping))
-	case *schema.ObjectType:
-		typ = mod.objectType(t, input, args)
-		if acceptMapping {
-			typ = fmt.Sprintf("pulumi.InputType[%s]", typ)
+	case *schema.OptionalType:
+		return fmt.Sprintf("Optional[%s]", mod.typeString(t.ElementType, input, acceptMapping))
+	case *schema.InputType:
+		typ := mod.typeString(codegen.SimplifyInputUnion(t.ElementType), input, acceptMapping)
+		if typ == "Any" {
+			return typ
 		}
+		return fmt.Sprintf("pulumi.Input[%s]", typ)
+	case *schema.EnumType:
+		return mod.tokenToEnum(t.Token)
+	case *schema.ArrayType:
+		return fmt.Sprintf("Sequence[%s]", mod.typeString(t.ElementType, input, acceptMapping))
+	case *schema.MapType:
+		return fmt.Sprintf("Mapping[str, %s]", mod.typeString(t.ElementType, input, acceptMapping))
+	case *schema.ObjectType:
+		typ := mod.objectType(t, input)
+		if !acceptMapping {
+			return typ
+		}
+		return fmt.Sprintf("pulumi.InputType[%s]", typ)
 	case *schema.ResourceType:
-		typ = fmt.Sprintf("'%s'", mod.resourceType(t))
+		return fmt.Sprintf("'%s'", mod.resourceType(t))
 	case *schema.TokenType:
 		// Use the underlying type for now.
 		if t.UnderlyingType != nil {
-			return mod.typeString(t.UnderlyingType, input, wrapInput, args, optional, acceptMapping)
+			return mod.typeString(t.UnderlyingType, input, acceptMapping)
 		}
-		typ = "Any"
+		return "Any"
 	case *schema.UnionType:
 		if !input {
 			for _, e := range t.ElementTypes {
 				// If this is an output and a "relaxed" enum, emit the type as the underlying primitive type rather than the union.
 				// Eg. Output[str] rather than Output[Any]
 				if typ, ok := e.(*schema.EnumType); ok {
-					return mod.typeString(typ.ElementType, input, wrapInput, args, optional, acceptMapping)
+					return mod.typeString(typ.ElementType, input, acceptMapping)
 				}
 			}
 			if t.DefaultType != nil {
-				return mod.typeString(t.DefaultType, input, wrapInput, args, optional, acceptMapping)
+				return mod.typeString(t.DefaultType, input, acceptMapping)
 			}
-			typ = "Any"
-		} else {
-			elementTypeSet := codegen.NewStringSet()
-			var elementTypes []schema.Type
-			for _, e := range t.ElementTypes {
-				et := mod.typeString(e, input, wrapInput, args, false, acceptMapping)
-				if !elementTypeSet.Has(et) {
-					elementTypeSet.Add(et)
-					elementTypes = append(elementTypes, e)
-				}
-			}
-
-			if len(elementTypes) == 1 {
-				return mod.typeString(elementTypes[0], input, wrapInput, args, optional, acceptMapping)
-			}
-
-			var elements []string
-			for _, e := range elementTypes {
-				t := mod.typeString(e, input, wrapInput, args, false, acceptMapping)
-				if args && strings.HasPrefix(t, "pulumi.Input[") {
-					contract.Assert(t[len(t)-1] == ']')
-					// Strip off the leading `pulumi.Input[` and the trailing `]`
-					t = t[len("pulumi.Input[") : len(t)-1]
-				}
-				elements = append(elements, t)
-			}
-			typ = fmt.Sprintf("Union[%s]", strings.Join(elements, ", "))
+			return "Any"
 		}
+
+		elementTypeSet := codegen.NewStringSet()
+		elements := make([]string, 0, len(t.ElementTypes))
+		for _, e := range t.ElementTypes {
+			et := mod.typeString(e, input, acceptMapping)
+			if !elementTypeSet.Has(et) {
+				elementTypeSet.Add(et)
+				elements = append(elements, et)
+			}
+		}
+
+		if len(elements) == 1 {
+			return elements[0]
+		}
+		return fmt.Sprintf("Union[%s]", strings.Join(elements, ", "))
 	default:
 		switch t {
 		case schema.BoolType:
-			typ = "bool"
+			return "bool"
 		case schema.IntType:
-			typ = "int"
+			return "int"
 		case schema.NumberType:
-			typ = "float"
+			return "float"
 		case schema.StringType:
-			typ = "str"
+			return "str"
 		case schema.ArchiveType:
-			typ = "pulumi.Archive"
+			return "pulumi.Archive"
 		case schema.AssetType:
-			typ = "Union[pulumi.Asset, pulumi.Archive]"
+			return "Union[pulumi.Asset, pulumi.Archive]"
 		case schema.JSONType:
 			fallthrough
 		case schema.AnyType:
-			typ = "Any"
+			return "Any"
 		}
 	}
 
-	if wrapInput && typ != "Any" {
-		typ = fmt.Sprintf("pulumi.Input[%s]", typ)
-	}
-	if optional {
-		return fmt.Sprintf("Optional[%s]", typ)
-	}
-	return typ
+	panic(fmt.Errorf("unexpected type %T", t))
 }
 
 // pyType returns the expected runtime type for the given variable.  Of course, being a dynamic language, this
 // check is not exhaustive, but it should be good enough to catch 80% of the cases early on.
 func (mod *modContext) pyType(typ schema.Type) string {
 	switch typ := typ.(type) {
+	case *schema.OptionalType:
+		return mod.pyType(typ.ElementType)
 	case *schema.EnumType:
 		return mod.pyType(typ.ElementType)
 	case *schema.ArrayType:
@@ -1999,6 +1994,8 @@ func (mod *modContext) pyType(typ schema.Type) string {
 }
 
 func isStringType(t schema.Type) bool {
+	t = codegen.UnwrapType(t)
+
 	for tt, ok := t.(*schema.TokenType); ok; tt, ok = t.(*schema.TokenType) {
 		t = tt.UnderlyingType
 	}
@@ -2028,35 +2025,20 @@ func InitParamName(name string) string {
 }
 
 func (mod *modContext) genObjectType(w io.Writer, obj *schema.ObjectType, input bool) error {
-	if input {
-		if mod.details(obj).argsType {
-			name := mod.unqualifiedObjectTypeName(obj, input, true)
-			if err := mod.genType(w, name, obj.Comment, obj.Properties, mod.details(obj).plainType, input, true, false); err != nil {
-				return err
-			}
-		}
-		if mod.details(obj).plainType {
-			name := mod.unqualifiedObjectTypeName(obj, input, false)
-			if err := mod.genType(w, name, obj.Comment, obj.Properties, mod.details(obj).plainType, input, false, false); err != nil {
-				return err
-			}
-		}
-		return nil
-	}
-
-	name := mod.unqualifiedObjectTypeName(obj, input, false)
-	return mod.genType(w, name, obj.Comment, obj.Properties, mod.details(obj).plainType, false, false, mod.details(obj).resourceOutputType)
+	name := mod.unqualifiedObjectTypeName(obj, input)
+	resourceOutputType := !input && mod.details(obj).resourceOutputType
+	return mod.genType(w, name, obj.Comment, obj.Properties, input, resourceOutputType)
 }
 
-func (mod *modContext) genType(w io.Writer, name, comment string, properties []*schema.Property, plainType, input, args, resourceOutput bool) error {
+func (mod *modContext) genType(w io.Writer, name, comment string, properties []*schema.Property, input, resourceOutput bool) error {
 	// Sort required props first.
 	props := make([]*schema.Property, len(properties))
 	copy(props, properties)
 	sort.Slice(props, func(i, j int) bool {
 		pi, pj := props[i], props[j]
 		switch {
-		case pi.IsRequired != pj.IsRequired:
-			return pi.IsRequired && !pj.IsRequired
+		case pi.IsRequired() != pj.IsRequired():
+			return pi.IsRequired() && !pj.IsRequired()
 		default:
 			return pi.Name < pj.Name
 		}
@@ -2126,15 +2108,15 @@ func (mod *modContext) genType(w io.Writer, name, comment string, properties []*
 	}
 	for _, prop := range props {
 		pname := PyName(prop.Name)
-		ty := mod.typeString(prop.Type, input, args && !prop.IsPlain, args && !prop.IsPlain, !prop.IsRequired, false /*acceptMapping*/)
+		ty := mod.typeString(prop.Type, input, false /*acceptMapping*/)
 		var defaultValue string
-		if !prop.IsRequired {
+		if !prop.IsRequired() {
 			defaultValue = " = None"
 		}
 		fmt.Fprintf(w, ",\n                 %s: %s%s", pname, ty, defaultValue)
 	}
 	fmt.Fprintf(w, "):\n")
-	mod.genTypeDocstring(w, comment, props, args)
+	mod.genTypeDocstring(w, comment, props)
 	if len(props) == 0 {
 		fmt.Fprintf(w, "        pass\n")
 	}
@@ -2145,7 +2127,7 @@ func (mod *modContext) genType(w io.Writer, name, comment string, properties []*
 
 		// Fill in computed defaults for arguments.
 		if prop.DefaultValue != nil {
-			dv, err := getDefaultValue(prop.DefaultValue, prop.Type)
+			dv, err := getDefaultValue(prop.DefaultValue, codegen.UnwrapType(prop.Type))
 			if err != nil {
 				return err
 			}
@@ -2172,7 +2154,7 @@ func (mod *modContext) genType(w io.Writer, name, comment string, properties []*
 		}
 
 		var indent string
-		if !prop.IsRequired {
+		if !prop.IsRequired() {
 			fmt.Fprintf(w, "        if %s is not None:\n", pname)
 			indent = "    "
 		}
@@ -2183,7 +2165,7 @@ func (mod *modContext) genType(w io.Writer, name, comment string, properties []*
 
 	// Generate properties. Input types have getters and setters, output types only have getters.
 	mod.genProperties(w, props, input /*setters*/, func(prop *schema.Property) string {
-		return mod.typeString(prop.Type, input, args && !prop.IsPlain, args && !prop.IsPlain, !prop.IsRequired, false /*acceptMapping*/)
+		return mod.typeString(prop.Type, input, false /*acceptMapping*/)
 	})
 
 	fmt.Fprintf(w, "\n")
@@ -2317,7 +2299,7 @@ func generateModuleContextMap(tool string, pkg *schema.Package, info PackageInfo
 		configMod.isConfig = true
 	}
 
-	visitObjectTypes(pkg.Config, func(t schema.Type, _ bool) {
+	visitObjectTypes(pkg.Config, func(t schema.Type) {
 		if t, ok := t.(*schema.ObjectType); ok {
 			getModFromToken(t.Token, t.Package).details(t).outputType = true
 		}
@@ -2327,33 +2309,24 @@ func generateModuleContextMap(tool string, pkg *schema.Package, info PackageInfo
 	scanResource := func(r *schema.Resource) {
 		mod := getModFromToken(r.Token, pkg)
 		mod.resources = append(mod.resources, r)
-		visitObjectTypes(r.Properties, func(t schema.Type, _ bool) {
+		visitObjectTypes(r.Properties, func(t schema.Type) {
 			switch T := t.(type) {
 			case *schema.ObjectType:
 				getModFromToken(T.Token, T.Package).details(T).outputType = true
 				getModFromToken(T.Token, T.Package).details(T).resourceOutputType = true
 			}
 		})
-		visitObjectTypes(r.InputProperties, func(t schema.Type, plain bool) {
+		visitObjectTypes(r.InputProperties, func(t schema.Type) {
 			switch T := t.(type) {
 			case *schema.ObjectType:
-				if r.IsProvider {
-					getModFromToken(T.Token, T.Package).details(T).outputType = true
-				}
 				getModFromToken(T.Token, T.Package).details(T).inputType = true
-				if plain {
-					getModFromToken(T.Token, T.Package).details(T).plainType = true
-				} else {
-					getModFromToken(T.Token, T.Package).details(T).argsType = true
-				}
 			}
 		})
 		if r.StateInputs != nil {
-			visitObjectTypes(r.StateInputs.Properties, func(t schema.Type, _ bool) {
+			visitObjectTypes(r.StateInputs.Properties, func(t schema.Type) {
 				switch T := t.(type) {
 				case *schema.ObjectType:
 					getModFromToken(T.Token, T.Package).details(T).inputType = true
-					getModFromToken(T.Token, T.Package).details(T).argsType = true
 				case *schema.ResourceType:
 					getModFromToken(T.Token, T.Resource.Package)
 				}
@@ -2371,7 +2344,7 @@ func generateModuleContextMap(tool string, pkg *schema.Package, info PackageInfo
 		mod := getModFromToken(f.Token, f.Package)
 		mod.functions = append(mod.functions, f)
 		if f.Inputs != nil {
-			visitObjectTypes(f.Inputs.Properties, func(t schema.Type, _ bool) {
+			visitObjectTypes(f.Inputs.Properties, func(t schema.Type) {
 				switch T := t.(type) {
 				case *schema.ObjectType:
 					getModFromToken(T.Token, T.Package).details(T).inputType = true
@@ -2382,7 +2355,7 @@ func generateModuleContextMap(tool string, pkg *schema.Package, info PackageInfo
 			})
 		}
 		if f.Outputs != nil {
-			visitObjectTypes(f.Outputs.Properties, func(t schema.Type, _ bool) {
+			visitObjectTypes(f.Outputs.Properties, func(t schema.Type) {
 				switch T := t.(type) {
 				case *schema.ObjectType:
 					getModFromToken(T.Token, T.Package).details(T).outputType = true
